@@ -10,28 +10,24 @@
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.
-"""A single HTTP request represented as a subclass of ``testtools.TestCase``
+"""A single HTTP request represented as a subclass of ``unittest.TestCase``
 
 The test case encapsulates the request headers and body and expected
 response headers and body. When the test is run an HTTP request is
-made using urllib3. Assertions are made against the reponse.
+made using httpx. Assertions are made against the response.
 """
 
 from collections import OrderedDict
 import copy
 import functools
+from http import cookies
 import os
 import re
 import sys
-from testtools import testcase
 import time
-from unittest import result
-
-import six
-from six.moves import http_cookies
-from six.moves.urllib import parse as urlparse
-import testtools
-import wsgi_intercept
+import unittest
+from unittest import result as unitresult
+import urllib.parse as urlparse
 
 from gabbi import __version__
 from gabbi import exception
@@ -53,11 +49,20 @@ REPLACERS = [
     'RESPONSE',
 ]
 
+# The list of approved type casts and associated functions
+APPROVED_CASTS = {
+    'int': int,
+    'float': float,
+    'str': str,
+    'bool': bool
+}
+
 # Basic test template determining both valid keys and default values
 BASE_TEST = {
     'name': '',
     'desc': '',
     'verbose': False,
+    'cert_validate': True,
     'ssl': False,
     'redirects': False,
     'method': 'GET',
@@ -69,6 +74,9 @@ BASE_TEST = {
     'xfail': False,
     'skip': '',
     'poll': {},
+    'use_prior_test': True,
+    'disable_response_handler': False,
+    'timeout': 30,
 }
 
 
@@ -80,21 +88,20 @@ def potentialFailure(func):
             try:
                 func(self)
             except Exception:
-                if hasattr(testcase, '_ExpectedFailure'):
-                    raise testcase._ExpectedFailure(sys.exc_info())
-                else:
-                    self._addExpectedFailure(self.result, sys.exc_info())
+                self._addExpectedFailure(self.result, sys.exc_info())
             else:
-                if hasattr(self, '_addUnexpectedSuccess'):
-                    self._addUnexpectedSuccess(self.result)
-                else:
-                    raise testcase._UnexpectedSuccess
+                self._addUnexpectedSuccess(self.result)
         else:
             func(self)
     return wrapper
 
 
-class HTTPTestCase(testtools.TestCase):
+def _is_complex_type(data):
+    """If data is a list or dict return True."""
+    return isinstance(data, list) or isinstance(data, dict)
+
+
+class HTTPTestCase(unittest.TestCase):
     """Encapsulate a single HTTP request as a TestCase.
 
     If the test is a member of a sequence of requests, ensure that prior
@@ -107,15 +114,26 @@ class HTTPTestCase(testtools.TestCase):
     base_test = copy.copy(BASE_TEST)
 
     def setUp(self):
+        self._fixture_cleanups = []
         if not self.has_run:
             super(HTTPTestCase, self).setUp()
             for fixture in self.inner_fixtures:
-                self.useFixture(fixture())
+                f = fixture()
+                f.setUp()
+                # add fixture to front of list so we unwind them FILO in
+                # tearDown
+                self._fixture_cleanups.insert(0, f)
 
     def tearDown(self):
         if not self.has_run:
             super(HTTPTestCase, self).tearDown()
         self.has_run = True
+        # Clean up an inner fixtures.
+        for fixture in self._fixture_cleanups:
+            fixture.cleanUp()
+
+    def shortDescription(self):
+        return self.test_base_name + ': ' + self.test_data['name']
 
     def run(self, result=None):
         """Store the current result handler on this test."""
@@ -128,16 +146,19 @@ class HTTPTestCase(testtools.TestCase):
 
         If there is a prior test in the sequence, run it first.
         """
+        self.cast = None
         if self.has_run:
             return
 
-        if self.test_data['skip']:
-            self.skipTest(self.test_data['skip'])
+        skip = self.replace_template(self.test_data['skip'])
+        if skip:
+            self.skipTest(skip)
 
-        if self.prior and not self.prior.has_run:
+        if (self.prior and not self.prior.has_run and
+                self.test_data['use_prior_test']):
             # Use a different result so we don't count this test
             # in the results.
-            self.prior.run(result.TestResult())
+            self.prior.run(unitresult.TestResult())
         self._run_test()
 
     def get_content_handler(self, content_type):
@@ -147,20 +168,26 @@ class HTTPTestCase(testtools.TestCase):
                 return handler
         return None
 
-    def replace_template(self, message):
+    def replace_template(self, message, escape_regex=False):
         """Replace magic strings in message."""
         if isinstance(message, dict):
             for k in message:
-                message[k] = self.replace_template(message[k])
+                message[k] = self.replace_template(message[k],
+                                                   escape_regex=escape_regex)
             return message
 
+        if isinstance(message, list):
+            return [self.replace_template(line, escape_regex=escape_regex)
+                    for line in message]
+
         for replacer in REPLACERS:
             template = '$%s' % replacer
             method = '_%s_replace' % replacer.lower()
             try:
                 if template in message:
                     try:
-                        message = getattr(self, method)(message)
+                        replace = getattr(self, method)
+                        message = replace(message, escape_regex=escape_regex)
                     except (KeyError, AttributeError, ValueError) as exc:
                         raise AssertionError(
                             'unable to replace %s in %s, data unavailable: %s'
@@ -171,10 +198,13 @@ class HTTPTestCase(testtools.TestCase):
 
         return message
 
+    def load_data_file(self, filename):
+        """Read a file from the current test directory."""
+        return self._load_data_file(filename)
+
     def _assert_response(self):
         """Compare the response with expected data."""
         self._test_status(self.test_data['status'], self.response['status'])
-
         for handler in self.response_handlers:
             handler(self)
 
@@ -186,32 +216,77 @@ class HTTPTestCase(testtools.TestCase):
         value = value.encode('UTF-8')
         return value
 
-    def _environ_replace(self, message):
-        """Replace an indicator in a message with the environment value."""
-        value = re.sub(self._replacer_regex('ENVIRON'),
-                       self._environ_replacer, message)
-        if value == "False":
-            return False
-        if value == "True":
-            return True
-        return value
-
     @staticmethod
-    def _environ_replacer(match):
+    def _regex_replacer(replacer, escape_regex):
+        """Wrap a replacer function to escape return values in a regex."""
+        if escape_regex:
+            @functools.wraps(replacer)
+            def replace(match):
+                return re.escape(replacer(match))
+
+            return replace
+        else:
+            return replacer
+
+    def _cast_value(self, value, message):
+        """Cast a replacement value the cast is in APPROVED_CASTS."""
+        # Turn off the casting flag, to reset state
+        cast = self.cast
+        self.cast = None
+        if cast in APPROVED_CASTS:
+            return APPROVED_CASTS[cast](value)
+        else:
+            raise RuntimeError('Invalid cast <%s> used in: %s' % (
+                cast, message))
+
+    def _environ_replace(self, message, escape_regex=False):
+        """Replace an indicator in a message with the environment value.
+
+        If value can be a number, cast it as such. If value is a form of
+        "null", "true", or "false" cast it to None, True, False.
+        """
+        value = re.sub(self._replacer_regex('ENVIRON'),
+                       self._regex_replacer(self._environ_replacer,
+                                            escape_regex),
+                       message)
+        if self.cast:
+            return self._cast_value(value, message)
+        else:
+            try:
+                if '.' in value:
+                    value = float(value)
+                else:
+                    value = int(value)
+                return value
+            except ValueError:
+                pass
+            if value.lower() == "false":
+                return False
+            if value.lower() == "true":
+                return True
+            if value.lower() == "null":
+                return None
+            return value
+
+    def _environ_replacer(self, match):
         """Replace a regex match with an environment value.
 
-        Let KeyError raise if variable not present.
+        Let KeyError raise if variable not present. Capture
+        the 'cast' if any.
         """
         environ_name = match.group('arg')
+        self.cast = match.group('cast')
         return os.environ[environ_name]
 
-    def _cookie_replace(self, message):
+    def _cookie_replace(self, message, escape_regex=False):
         """Replace $COOKIE in a message.
 
         With cookie data from set-cookie in the prior request.
         """
         return re.sub(self._simple_replacer_regex('COOKIE'),
-                      self._cookie_replacer, message)
+                      self._regex_replacer(self._cookie_replacer,
+                                           escape_regex),
+                      message)
 
     def _cookie_replacer(self, match):
         """Replace a regex match with the cookie of a previous response."""
@@ -220,18 +295,21 @@ class HTTPTestCase(testtools.TestCase):
             referred_case = self.history[case]
         else:
             referred_case = self.prior
-        response_cookies = referred_case.response['set-cookie']
-        cookies = http_cookies.SimpleCookie()
-        cookies.load(response_cookies)
-        cookie_string = cookies.output(attrs=[], header='', sep=',').strip()
+        response_cookie = referred_case.response['set-cookie']
+        cookie = cookies.SimpleCookie()
+        cookie.load(response_cookie)
+        cookie_string = cookie.output(attrs=[], header='', sep=',').strip()
         return cookie_string
 
-    def _headers_replace(self, message):
-        """Replace a header indicator in a message with that headers value from
-        the prior request.
+    def _headers_replace(self, message, escape_regex=False):
+        """Replace a header indicator in a message.
+
+        Replace it with the header's value from the prior request.
         """
         return re.sub(self._replacer_regex('HEADERS'),
-                      self._header_replacer, message)
+                      self._regex_replacer(self._header_replacer,
+                                           escape_regex),
+                      message)
 
     def _header_replacer(self, match):
         """Replace a regex match with the value of a prior header."""
@@ -243,20 +321,25 @@ class HTTPTestCase(testtools.TestCase):
             referred_case = self.prior
         return referred_case.response[header_key.lower()]
 
-    def _last_url_replace(self, message):
+    def _last_url_replace(self, message, escape_regex=False):
         """Replace $LAST_URL in a message.
 
         With the URL used in the prior request.
         """
-        return message.replace('$LAST_URL', self.prior.url)
+        last_url = self.prior.url
+        if escape_regex:
+            last_url = re.escape(last_url)
+        return message.replace('$LAST_URL', last_url)
 
-    def _url_replace(self, message):
+    def _url_replace(self, message, escape_regex=False):
         """Replace $URL in a message.
 
         With the URL used in a previous request.
         """
         return re.sub(self._simple_replacer_regex('URL'),
-                      self._url_replacer, message)
+                      self._regex_replacer(self._url_replacer,
+                                           escape_regex),
+                      message)
 
     def _url_replacer(self, match):
         """Replace a regex match with the value of a previous url."""
@@ -267,13 +350,15 @@ class HTTPTestCase(testtools.TestCase):
             referred_case = self.prior
         return referred_case.url
 
-    def _location_replace(self, message):
+    def _location_replace(self, message, escape_regex=False):
         """Replace $LOCATION in a message.
 
         With the location header from a previous request.
         """
         return re.sub(self._simple_replacer_regex('LOCATION'),
-                      self._location_replacer, message)
+                      self._regex_replacer(self._location_replacer,
+                                           escape_regex),
+                      message)
 
     def _location_replacer(self, match):
         """Replace a regex match with the value of a previous location."""
@@ -296,11 +381,13 @@ class HTTPTestCase(testtools.TestCase):
         with open(path, mode='rb') as data_file:
             return data_file.read()
 
-    def _netloc_replace(self, message):
+    def _netloc_replace(self, message, escape_regex=False):
         """Replace $NETLOC with the current host and port."""
         netloc = self.netloc
         if self.prefix:
             netloc = '%s%s' % (netloc, self.prefix)
+        if escape_regex:
+            netloc = re.escape(netloc)
         return message.replace('$NETLOC', netloc)
 
     def _parse_url(self, url):
@@ -341,8 +428,10 @@ class HTTPTestCase(testtools.TestCase):
     def _replacer_regex(key):
         """Compose a regular expression for test template variables."""
         case = HTTPTestCase._history_regex
-        return r"%s\$%s\[(?P<quote>['\"])(?P<arg>.+?)(?P=quote)\]" % (
-            case, key)
+        return (
+            r"%s\$%s(:(?P<cast>\w+))?"
+            r"\[(?P<quote>['\"])(?P<arg>.+?)(?P=quote)\]"
+            % (case, key))
 
     @staticmethod
     def _simple_replacer_regex(key):
@@ -350,15 +439,30 @@ class HTTPTestCase(testtools.TestCase):
         case = HTTPTestCase._history_regex
         return r"%s\$%s" % (case, key)
 
-    def _response_replace(self, message):
-        """Replace a content path with the value from a previous response."""
-        return re.sub(self._replacer_regex('RESPONSE'),
-                      self._response_replacer, message)
+    def _response_replace(self, message, escape_regex=False):
+        """Replace a content path with the value from a previous response.
 
-    def _response_replacer(self, match):
+        If the match would replace the entire message, then don't cast it
+        to a string.
+        """
+        regex = self._replacer_regex('RESPONSE')
+        match = re.match('^%s$' % regex, message)
+        if match:
+            return self._response_replacer(match, preserve=True)
+        return re.sub(regex,
+                      self._regex_replacer(self._response_replacer,
+                                           escape_regex),
+                      message)
+
+    def _response_replacer(self, match, preserve=False):
         """Replace a regex match with the value from a previous response."""
         response_path = match.group('arg')
         case = match.group('case')
+        self.cast = match.group('cast')
+        if not preserve and self.cast:
+            raise RuntimeError(
+                'Unable to process cast <%s> within message: %s'
+                % (self.cast, match.string))
         if case:
             referred_case = self.history[case]
         else:
@@ -368,32 +472,37 @@ class HTTPTestCase(testtools.TestCase):
         # If no handler can be found use the null replacer,
         # which returns "foo" when "$RESPONSE['foo']".
         replacer_class = replacer_class or base.ContentHandler
-        return replacer_class.replacer(
+        result = replacer_class.replacer(
             referred_case.response_data, response_path)
+        if preserve:
+            if self.cast:
+                return self._cast_value(result, match.string)
+            return result
+        else:
+            return str(result)
 
-    def _run_request(self, url, method, headers, body, redirect=False):
-        """Run the http request and decode output.
-
-        The call to make the request will catch a WSGIAppError from
-        wsgi_intercept so that the real traceback from a catastrophic
-        error in the intercepted app can be examined.
-        """
+    def _run_request(
+        self,
+        url,
+        method,
+        headers,
+        body,
+        redirect=False,
+        timeout=30,
+    ):
+        """Run the http request and decode output."""
 
         if 'user-agent' not in (key.lower() for key in headers):
-            headers['user-agent'] = "gabbi/%s (Python urllib3)" % __version__
+            headers['user-agent'] = "gabbi/%s (Python httpx)" % __version__
 
-        try:
-            response, content = self.http.request(
-                url,
-                method=method,
-                headers=headers,
-                body=body,
-                redirect=redirect
-            )
-        except wsgi_intercept.WSGIAppError as exc:
-            # Extract and re-raise the wrapped exception.
-            six.reraise(exc.exception_type, exc.exception_value,
-                        exc.traceback)
+        response, content = self.http.request(
+            url,
+            method=method,
+            headers=headers,
+            body=body,
+            redirect=redirect,
+            timeout=timeout,
+        )
 
         # Set headers and location attributes for follow on requests
         self.response = response
@@ -404,13 +513,34 @@ class HTTPTestCase(testtools.TestCase):
         decoded_output = utils.decode_response_content(response, content)
         self.content_type = response.get('content-type', '').lower()
         loader_class = self.get_content_handler(self.content_type)
-        if decoded_output and loader_class:
+        if (decoded_output and loader_class
+                and not self.test_data['disable_response_handler']):
             # save structured response data
-            self.response_data = loader_class.loads(decoded_output)
+            try:
+                self.response_data = loader_class.loads(decoded_output)
+            except exception.GabbiDataLoadError as exc:
+                raise AssertionError(
+                    'unable to load data as %s' % self.content_type) from exc
+
         else:
             self.response_data = None
         self.output = decoded_output
 
+    def _replace_headers_template(self, test_name, headers):
+        replaced_headers = {}
+
+        try:
+            for name in headers:
+                replaced_name = self.replace_template(name)
+                replaced_headers[replaced_name] = self.replace_template(
+                    headers[name]
+                )
+        except TypeError as exc:
+            raise exception.GabbiFormatError(
+                'malformed headers in test %s: %s' % (test_name, exc))
+
+        return replaced_headers
+
     def _run_test(self):
         """Make an HTTP request and compare the response with expectations."""
         test = self.test_data
@@ -420,14 +550,15 @@ class HTTPTestCase(testtools.TestCase):
         self.url = base_url
         full_url = self._parse_url(base_url)
 
+        # Replace variables in headers with variable values. This includes both
+        # in the header key and the header value.
+        test['request_headers'] = self._replace_headers_template(
+            test['name'], test['request_headers'])
+        test['response_headers'] = self._replace_headers_template(
+            test['name'], test['response_headers'])
+
         method = test['method'].upper()
         headers = test['request_headers']
-        for name in headers:
-            try:
-                headers[name] = self.replace_template(headers[name])
-            except TypeError as exc:
-                raise exception.GabbiFormatError(
-                    'malformed headers in test %s: %s' % (test['name'], exc))
 
         if test['data'] != '':
             body = self._test_data_to_string(
@@ -436,14 +567,26 @@ class HTTPTestCase(testtools.TestCase):
         else:
             body = ''
 
+        # ensure body is bytes, encoding as UTF-8 because that's
+        # what we do here
+        if isinstance(body, str):
+            body = body.encode('UTF-8')
+
         if test['poll']:
-            count = test['poll'].get('count', 1)
-            delay = test['poll'].get('delay', 1)
+            count = int(float(self.replace_template(
+                test['poll'].get('count', 1))))
+            delay = float(self.replace_template(test['poll'].get('delay', 1)))
             failure = None
             while count:
                 try:
-                    self._run_request(full_url, method, headers, body,
-                                      redirect=test['redirects'])
+                    self._run_request(
+                        full_url,
+                        method,
+                        headers,
+                        body,
+                        redirect=test['redirects'],
+                        timeout=test['timeout'],
+                    )
                     self._assert_response()
                     failure = None
                     break
@@ -456,38 +599,60 @@ class HTTPTestCase(testtools.TestCase):
             if failure:
                 raise failure
         else:
-            self._run_request(full_url, method, headers, body,
-                              redirect=test['redirects'])
+            self._run_request(
+                full_url,
+                method,
+                headers,
+                body,
+                redirect=test['redirects'],
+                timeout=test['timeout'],
+            )
             self._assert_response()
 
-    def _scheme_replace(self, message):
+    def _scheme_replace(self, message, escape_regex=False):
         """Replace $SCHEME with the current protocol."""
-        return message.replace('$SCHEME', self.scheme)
+        scheme = re.escape(self.scheme) if escape_regex else self.scheme
+        return message.replace('$SCHEME', scheme)
 
     def _test_data_to_string(self, data, content_type):
         """Turn the request data into a string.
 
         If the data is not binary, replace template strings.
+
+        If the result of the template handling is not a string,
+        run the result through the dumper.
         """
-        if isinstance(data, str):
-            if data.startswith('<@'):
-                info = self._load_data_file(data.replace('<@', '', 1))
+        dumper_class = self.get_content_handler(content_type)
+        if not _is_complex_type(data):
+            if isinstance(data, str) and data.startswith('<@'):
+                info = self.load_data_file(data.replace('<@', '', 1))
                 if utils.not_binary(content_type):
-                    try:
-                        info = str(info, 'UTF-8')
-                    except TypeError:
-                        info = info.encode('UTF-8')
-                    data = info
+                    data = str(info, 'UTF-8')
                 else:
+                    # Return early we are binary content
                     return info
         else:
-            dumper_class = self.get_content_handler(content_type)
+            # We have a complex data structure, try to dump it.
             if dumper_class:
-                data = dumper_class.dumps(data)
+                data = self.replace_template(data)
+                data = dumper_class.dumps(data, test=self)
             else:
-                raise ValueError(
-                    'unable to process data to %s' % content_type)
-        return self.replace_template(data)
+                if content_type:
+                    raise ValueError(
+                        'unable to process data to %s' % content_type)
+                else:
+                    raise ValueError(
+                        'no content-type available for processing data')
+
+        data = self.replace_template(data)
+
+        # If the result after template handling is not a string, dump
+        # it if there is a suitable dumper.
+        if dumper_class and not isinstance(data, str):
+            # If there are errors dumping we want them to raise to the
+            # test harness.
+            data = dumper_class.dumps(data, test=self)
+        return data
 
     def _test_status(self, expected_status, observed_status):
         """Confirm we got the expected status.
@@ -542,7 +707,7 @@ class HTTPTestCase(testtools.TestCase):
                 dumper_class = self.get_content_handler(self.content_type)
                 if dumper_class:
                     full_response = dumper_class.dumps(self.response_data,
-                                                       pretty=True)
+                                                       pretty=True, test=self)
                 else:
                     full_response = self.output
             else:
